@@ -1,24 +1,24 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
  *   Free Software Foundation; either version 2 of the License, or (at your
  *   option) any later version.
- * 
+ *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  *   See the GNU General Public License for more details.
- * 
+ *
  *   You should have received a copy of the GNU General Public License along
  *   with this program; if not, see <http://www.gnu.org/licenses>.
- * 
+ *
  *   Linking stunnel statically or dynamically with other modules is making
  *   a combined work based on stunnel. Thus, the terms and conditions of
  *   the GNU General Public License cover the whole combination.
- * 
+ *
  *   In addition, as a special exception, the copyright holder of stunnel
  *   gives you permission to combine stunnel with free software programs or
  *   libraries that are released under the GNU LGPL and with code included
@@ -26,7 +26,7 @@
  *   modified versions of such code, with unchanged license). You may copy
  *   and distribute such a system following the terms of the GNU GPL for
  *   stunnel and the licenses of the other code concerned.
- * 
+ *
  *   Note that people who make modified versions of stunnel are not obligated
  *   to grant this special exception for their modified versions; it is their
  *   choice whether to do so. The GNU General Public License gives permission
@@ -38,31 +38,31 @@
 #include "common.h"
 #include "prototypes.h"
 
-#define isprefix(a, b) (strncasecmp((a), (b), strlen(b))==0)
+#define is_prefix(a, b) (strncasecmp((a), (b), strlen(b))==0)
 
 /* protocol-specific function prototypes */
-static void proxy_server(CLI *c);
-static void cifs_client(CLI *);
-static void cifs_server(CLI *);
-static void pgsql_client(CLI *);
-static void pgsql_server(CLI *);
-static void smtp_client(CLI *);
-static void smtp_server(CLI *);
-static void pop3_client(CLI *);
-static void pop3_server(CLI *);
-static void imap_client(CLI *);
-static void imap_server(CLI *);
-static void nntp_client(CLI *);
-static void connect_server(CLI *);
-static void connect_client(CLI *);
+NOEXPORT void proxy_server(CLI *c);
+NOEXPORT void cifs_client(CLI *);
+NOEXPORT void cifs_server(CLI *);
+NOEXPORT void pgsql_client(CLI *);
+NOEXPORT void pgsql_server(CLI *);
+NOEXPORT void smtp_client(CLI *);
+NOEXPORT void smtp_server(CLI *);
+NOEXPORT void pop3_client(CLI *);
+NOEXPORT void pop3_server(CLI *);
+NOEXPORT void imap_client(CLI *);
+NOEXPORT void imap_server(CLI *);
+NOEXPORT void nntp_client(CLI *);
+NOEXPORT void connect_server(CLI *);
+NOEXPORT void connect_client(CLI *);
 
 #if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
-static void ntlm(CLI *);
-static char *ntlm1();
-static char *ntlm3(char *, char *, char *);
-static void crypt_DES(DES_cblock, DES_cblock, DES_cblock);
+NOEXPORT void ntlm(CLI *);
+NOEXPORT char *ntlm1();
+NOEXPORT char *ntlm3(char *, char *, char *);
+NOEXPORT void crypt_DES(DES_cblock, DES_cblock, DES_cblock);
 #endif
-static char *base64(int, char *, int);
+NOEXPORT char *base64(int, char *, int);
 
 /**************************************** framework */
 
@@ -71,7 +71,7 @@ typedef void (*FUNCTION)(CLI *);
 static const struct {
     char *name;
     struct {
-        PROTOCOL_TYPE type;
+        PROTOCOL_PHASE type;
         FUNCTION func;
     } handlers[2];
 } protocols[]={
@@ -95,7 +95,7 @@ int find_protocol_id(const char *name) {
     return -1;
 }
 
-void protocol(CLI *c, const PROTOCOL_TYPE type) {
+void protocol(CLI *c, const PROTOCOL_PHASE type) {
     const int id=c->opt->protocol, mode=(unsigned int)c->opt->option.client;
 
     if(id<0 || type!=protocols[id].handlers[mode].type ||
@@ -112,7 +112,7 @@ void protocol(CLI *c, const PROTOCOL_TYPE type) {
 
 /*
  * PROXY protocol: http://haproxy.1wt.eu/download/1.5/doc/proxy-protocol.txt
- * this is a protocol client support for stunnel acting as an SSL server 
+ * this is a protocol client support for stunnel acting as an SSL server
  * I don't think anything else is useful, but feel free to discuss on the
  * stunnel-users mailing list if you disagree
  */
@@ -122,7 +122,7 @@ void protocol(CLI *c, const PROTOCOL_TYPE type) {
 #define IP_LEN 40
 #define PORT_LEN 6
 
-static void proxy_server(CLI *c) {
+NOEXPORT void proxy_server(CLI *c) {
     SOCKADDR_UNION addr;
     socklen_t addrlen;
     char src_host[IP_LEN], dst_host[IP_LEN];
@@ -157,9 +157,11 @@ static void proxy_server(CLI *c) {
     case AF_INET:
         proto="TCP4";
         break;
+#ifdef USE_IPv6
     case AF_INET6:
         proto="TCP6";
         break;
+#endif
     default: /* AF_UNIX */
         proto="UNKNOWN";
     }
@@ -169,12 +171,12 @@ static void proxy_server(CLI *c) {
 
 /**************************************** cifs */
 
-static void cifs_client(CLI *c) {
+NOEXPORT void cifs_client(CLI *c) {
     u8 buffer[5];
     u8 request_dummy[4] = {0x81, 0, 0, 0}; /* a zero-length request */
 
-    write_blocking(c, c->remote_fd.fd, request_dummy, 4);
-    read_blocking(c, c->remote_fd.fd, buffer, 5);
+    s_write(c, c->remote_fd.fd, request_dummy, 4);
+    s_read(c, c->remote_fd.fd, buffer, 5);
     if(buffer[0]!=0x83) { /* NB_SSN_NEGRESP */
         s_log(LOG_ERR, "Negative response expected");
         longjmp(c->err, 1);
@@ -189,26 +191,26 @@ static void cifs_client(CLI *c) {
     }
 }
 
-static void cifs_server(CLI *c) {
+NOEXPORT void cifs_server(CLI *c) {
     u8 buffer[128];
     u8 response_access_denied[5] = {0x83, 0, 0, 1, 0x81};
     u8 response_use_ssl[5] = {0x83, 0, 0, 1, 0x8e};
     u16 len;
 
-    read_blocking(c, c->local_rfd.fd, buffer, 4) ;/* NetBIOS header */
+    s_read(c, c->local_rfd.fd, buffer, 4) ;/* NetBIOS header */
     len=buffer[3];
     len|=(u16)(buffer[2]) << 8;
     if(len>sizeof buffer-4) {
         s_log(LOG_ERR, "Received block too long");
         longjmp(c->err, 1);
     }
-    read_blocking(c, c->local_rfd.fd, buffer+4, len);
-    if(buffer[0]!=0x81){ /* NB_SSN_REQUEST */
+    s_read(c, c->local_rfd.fd, buffer+4, len);
+    if(buffer[0]!=0x81) { /* NB_SSN_REQUEST */
         s_log(LOG_ERR, "Client did not send session setup");
-        write_blocking(c, c->local_wfd.fd, response_access_denied, 5);
+        s_write(c, c->local_wfd.fd, response_access_denied, 5);
         longjmp(c->err, 1);
     }
-    write_blocking(c, c->local_wfd.fd, response_use_ssl, 5);
+    s_write(c, c->local_wfd.fd, response_use_ssl, 5);
 }
 
 /**************************************** pgsql */
@@ -216,11 +218,11 @@ static void cifs_server(CLI *c) {
 /* http://www.postgresql.org/docs/8.3/static/protocol-flow.html#AEN73982 */
 u8 ssl_request[8]={0, 0, 0, 8, 0x04, 0xd2, 0x16, 0x2f};
 
-static void pgsql_client(CLI *c) {
+NOEXPORT void pgsql_client(CLI *c) {
     u8 buffer[1];
 
-    write_blocking(c, c->remote_fd.fd, ssl_request, sizeof ssl_request);
-    read_blocking(c, c->remote_fd.fd, buffer, 1);
+    s_write(c, c->remote_fd.fd, ssl_request, sizeof ssl_request);
+    s_read(c, c->remote_fd.fd, buffer, 1);
     /* S - accepted, N - rejected, non-SSL preferred */
     if(buffer[0]!='S') {
         s_log(LOG_ERR, "PostgreSQL server rejected SSL");
@@ -228,49 +230,56 @@ static void pgsql_client(CLI *c) {
     }
 }
 
-static void pgsql_server(CLI *c) {
+NOEXPORT void pgsql_server(CLI *c) {
     u8 buffer[8], ssl_ok[1]={'S'};
 
     memset(buffer, 0, sizeof buffer);
-    read_blocking(c, c->local_rfd.fd, buffer, sizeof buffer);
+    s_read(c, c->local_rfd.fd, buffer, sizeof buffer);
     if(memcmp(buffer, ssl_request, sizeof ssl_request)) {
         s_log(LOG_ERR, "PostgreSQL client did not request SSL, rejecting");
         /* no way to send error on startup, so just drop the client */
         longjmp(c->err, 1);
     }
-    write_blocking(c, c->local_wfd.fd, ssl_ok, sizeof ssl_ok);
+    s_write(c, c->local_wfd.fd, ssl_ok, sizeof ssl_ok);
 }
 
 /**************************************** smtp */
 
-static void smtp_client(CLI *c) {
+NOEXPORT void smtp_client(CLI *c) {
     char *line;
 
+    line=str_dup("");
     do { /* copy multiline greeting */
+        str_free(line);
         line=fd_getline(c, c->remote_fd.fd);
         fd_putline(c, c->local_wfd.fd, line);
-    } while(isprefix(line, "220-"));
+    } while(is_prefix(line, "220-"));
 
     fd_putline(c, c->remote_fd.fd, "EHLO localhost");
     do { /* skip multiline reply */
+        str_free(line);
         line=fd_getline(c, c->remote_fd.fd);
-    } while(isprefix(line, "250-"));
-    if(!isprefix(line, "250 ")) { /* error */
+    } while(is_prefix(line, "250-"));
+    if(!is_prefix(line, "250 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 1425 compliant");
+        str_free(line);
         longjmp(c->err, 1);
     }
 
     fd_putline(c, c->remote_fd.fd, "STARTTLS");
     do { /* skip multiline reply */
+        str_free(line);
         line=fd_getline(c, c->remote_fd.fd);
-    } while(isprefix(line, "220-"));
-    if(!isprefix(line, "220 ")) { /* error */
+    } while(is_prefix(line, "220-"));
+    if(!is_prefix(line, "220 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 2487 compliant");
+        str_free(line);
         longjmp(c->err, 1);
     }
+    str_free(line);
 }
 
-static void smtp_server(CLI *c) {
+NOEXPORT void smtp_server(CLI *c) {
     char *line;
 
     s_poll_init(c->fds);
@@ -288,88 +297,106 @@ static void smtp_server(CLI *c) {
     }
 
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "220")) {
+    if(!is_prefix(line, "220")) {
         s_log(LOG_ERR, "Unknown server welcome");
+        str_free(line);
         longjmp(c->err, 1);
     }
     fd_printf(c, c->local_wfd.fd, "%s + stunnel", line);
+    str_free(line);
     line=fd_getline(c, c->local_rfd.fd);
-    if(!isprefix(line, "EHLO ")) {
+    if(!is_prefix(line, "EHLO ")) {
         s_log(LOG_ERR, "Unknown client EHLO");
+        str_free(line);
         longjmp(c->err, 1);
     }
     fd_printf(c, c->local_wfd.fd, "250-%s Welcome", line);
     fd_putline(c, c->local_wfd.fd, "250 STARTTLS");
+    str_free(line);
     line=fd_getline(c, c->local_rfd.fd);
-    if(!isprefix(line, "STARTTLS")) {
+    if(!is_prefix(line, "STARTTLS")) {
         s_log(LOG_ERR, "STARTTLS expected");
+        str_free(line);
         longjmp(c->err, 1);
     }
     fd_putline(c, c->local_wfd.fd, "220 Go ahead");
+    str_free(line);
 }
 
 /**************************************** pop3 */
 
-static void pop3_client(CLI *c) {
+NOEXPORT void pop3_client(CLI *c) {
     char *line;
 
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "+OK ")) {
+    if(!is_prefix(line, "+OK ")) {
         s_log(LOG_ERR, "Unknown server welcome");
+        str_free(line);
         longjmp(c->err, 1);
     }
     fd_putline(c, c->local_wfd.fd, line);
     fd_putline(c, c->remote_fd.fd, "STLS");
+    str_free(line);
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "+OK ")) {
+    if(!is_prefix(line, "+OK ")) {
         s_log(LOG_ERR, "Server does not support TLS");
+        str_free(line);
         longjmp(c->err, 1);
     }
+    str_free(line);
 }
 
-static void pop3_server(CLI *c) {
+NOEXPORT void pop3_server(CLI *c) {
     char *line;
 
     line=fd_getline(c, c->remote_fd.fd);
     fd_printf(c, c->local_wfd.fd, "%s + stunnel", line);
+    str_free(line);
     line=fd_getline(c, c->local_rfd.fd);
-    if(isprefix(line, "CAPA")) { /* client wants RFC 2449 extensions */
+    if(is_prefix(line, "CAPA")) { /* client wants RFC 2449 extensions */
         fd_putline(c, c->local_wfd.fd, "+OK Stunnel capability list follows");
         fd_putline(c, c->local_wfd.fd, "STLS");
         fd_putline(c, c->local_wfd.fd, ".");
+        str_free(line);
         line=fd_getline(c, c->local_rfd.fd);
     }
-    if(!isprefix(line, "STLS")) {
+    if(!is_prefix(line, "STLS")) {
         s_log(LOG_ERR, "Client does not want TLS");
+        str_free(line);
         longjmp(c->err, 1);
     }
+    str_free(line);
     fd_putline(c, c->local_wfd.fd, "+OK Stunnel starts TLS negotiation");
 }
 
 /**************************************** imap */
 
-static void imap_client(CLI *c) {
+NOEXPORT void imap_client(CLI *c) {
     char *line;
 
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "* OK")) {
+    if(!is_prefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
+        str_free(line);
         longjmp(c->err, 1);
     }
     fd_putline(c, c->local_wfd.fd, line);
     fd_putline(c, c->remote_fd.fd, "stunnel STARTTLS");
+    str_free(line);
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "stunnel OK")) {
+    if(!is_prefix(line, "stunnel OK")) {
         fd_putline(c, c->local_wfd.fd,
             "* BYE stunnel: Server does not support TLS");
         s_log(LOG_ERR, "Server does not support TLS");
+        str_free(line);
         longjmp(c->err, 2); /* don't reset */
     }
+    str_free(line);
 }
 
-static void imap_server(CLI *c) {
+NOEXPORT void imap_server(CLI *c) {
     char *line, *id, *tail, *capa;
- 
+
     s_poll_init(c->fds);
     s_poll_add(c->fds, c->local_rfd.fd, 1, 0);
     switch(s_poll_wait(c->fds, 0, 200)) {
@@ -386,8 +413,9 @@ static void imap_server(CLI *c) {
 
     /* process server welcome and send it to client */
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "* OK")) {
+    if(!is_prefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
+        str_free(line);
         longjmp(c->err, 1);
     }
     capa=strstr(line, "CAPABILITY");
@@ -397,41 +425,48 @@ static void imap_server(CLI *c) {
         *capa='K'; /* disable CAPABILITY within greeting */
     fd_printf(c, c->local_wfd.fd, "%s (stunnel)", line);
 
+    id=str_dup("");
     while(1) { /* process client commands */
+        str_free(line);
         line=fd_getline(c, c->local_rfd.fd);
         /* split line into id and tail */
+        str_free(id);
         id=str_dup(line);
         tail=strchr(id, ' ');
         if(!tail)
             break;
         *tail++='\0';
 
-        if(isprefix(tail, "STARTTLS")) {
+        if(is_prefix(tail, "STARTTLS")) {
             fd_printf(c, c->local_wfd.fd,
                 "%s OK Begin TLS negotiation now", id);
+            str_free(line);
+            str_free(id);
             return; /* success */
-        } else if(isprefix(tail, "CAPABILITY")) {
+        } else if(is_prefix(tail, "CAPABILITY")) {
             fd_putline(c, c->remote_fd.fd, line); /* send it to server */
+            str_free(line);
             line=fd_getline(c, c->remote_fd.fd); /* get the capabilites */
             if(*line=='*') {
-                /* 
+                /*
                  * append STARTTLS
                  * should also add LOGINDISABLED, but can't because
                  * of Mozilla bug #324138/#312009
                  * LOGIN would fail as "unexpected command", anyway
                  */
                 fd_printf(c, c->local_wfd.fd, "%s STARTTLS", line);
+                str_free(line);
                 line=fd_getline(c, c->remote_fd.fd); /* next line */
             }
             fd_putline(c, c->local_wfd.fd, line); /* forward to the client */
             tail=strchr(line, ' ');
-            if(!tail || !isprefix(tail+1, "OK")) { /* not OK? */
+            if(!tail || !is_prefix(tail+1, "OK")) { /* not OK? */
                 fd_putline(c, c->local_wfd.fd,
                     "* BYE unexpected server response");
                 s_log(LOG_ERR, "Unexpected server response: %s", line);
                 break;
             }
-        } else if(isprefix(tail, "LOGOUT")) {
+        } else if(is_prefix(tail, "LOGOUT")) {
             fd_putline(c, c->local_wfd.fd, "* BYE server terminating");
             fd_printf(c, c->local_wfd.fd, "%s OK LOGOUT completed", id);
             break;
@@ -441,64 +476,81 @@ static void imap_server(CLI *c) {
             s_log(LOG_ERR, "Unexpected client command %s", tail);
             break;
         }
-    }    
+    }
     /* clean server shutdown */
+    str_free(id);
     fd_putline(c, c->remote_fd.fd, "stunnel LOGOUT");
+    str_free(line);
     line=fd_getline(c, c->remote_fd.fd);
-    if(*line=='*')
+    if(*line=='*') {
+        str_free(line);
         line=fd_getline(c, c->remote_fd.fd);
+    }
+    str_free(line);
     longjmp(c->err, 2); /* don't reset */
 }
 
 /**************************************** nntp */
 
-static void nntp_client(CLI *c) {
+NOEXPORT void nntp_client(CLI *c) {
     char *line;
 
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "200 ") && !isprefix(line, "201 ")) {
+    if(!is_prefix(line, "200 ") && !is_prefix(line, "201 ")) {
         s_log(LOG_ERR, "Unknown server welcome");
+        str_free(line);
         longjmp(c->err, 1);
     }
     fd_putline(c, c->local_wfd.fd, line);
     fd_putline(c, c->remote_fd.fd, "STARTTLS");
+    str_free(line);
     line=fd_getline(c, c->remote_fd.fd);
-    if(!isprefix(line, "382 ")) {
+    if(!is_prefix(line, "382 ")) {
         s_log(LOG_ERR, "Server does not support TLS");
+        str_free(line);
         longjmp(c->err, 1);
     }
+    str_free(line);
 }
 
 /**************************************** connect */
 
-static void connect_server(CLI *c) {
+NOEXPORT void connect_server(CLI *c) {
     char *request, *proto, *header;
-    int not_empty;
+    NAME_LIST host_list;
 
     request=fd_getline(c, c->local_rfd.fd);
-    if(!isprefix(request, "CONNECT ")) {
+    if(!is_prefix(request, "CONNECT ")) {
         fd_putline(c, c->local_wfd.fd, "HTTP/1.0 400 Bad Request Method");
         fd_putline(c, c->local_wfd.fd, "Server: stunnel/" STUNNEL_VERSION);
         fd_putline(c, c->local_wfd.fd, "");
+        str_free(request);
         longjmp(c->err, 1);
     }
     proto=strchr(request+8, ' ');
-    if(!proto || !isprefix(proto, " HTTP/")) {
+    if(!proto || !is_prefix(proto, " HTTP/")) {
         fd_putline(c, c->local_wfd.fd, "HTTP/1.0 400 Bad Request Protocol");
         fd_putline(c, c->local_wfd.fd, "Server: stunnel/" STUNNEL_VERSION);
         fd_putline(c, c->local_wfd.fd, "");
+        str_free(request);
         longjmp(c->err, 1);
     }
     *proto='\0';
-    do { /* ignore any headers*/
-        header=fd_getline(c, c->local_rfd.fd);
-        not_empty=*header;
+
+    header=str_dup("");
+    do { /* ignore any headers */
         str_free(header);
-    } while(not_empty);
-    if(!name2addrlist(&c->connect_addr, request+8, DEFAULT_LOOPBACK)) {
+        header=fd_getline(c, c->local_rfd.fd);
+    } while(*header); /* not empty */
+    str_free(header);
+
+    host_list.name=request+8;
+    host_list.next=NULL;
+    if(!namelist2addrlist(&c->connect_addr, &host_list, DEFAULT_LOOPBACK)) {
         fd_putline(c, c->local_wfd.fd, "HTTP/1.0 404 Not Found");
         fd_putline(c, c->local_wfd.fd, "Server: stunnel/" STUNNEL_VERSION);
         fd_putline(c, c->local_wfd.fd, "");
+        str_free(request);
         longjmp(c->err, 1);
     }
     str_free(request);
@@ -507,7 +559,7 @@ static void connect_server(CLI *c) {
     fd_putline(c, c->local_wfd.fd, "");
 }
 
-static void connect_client(CLI *c) {
+NOEXPORT void connect_client(CLI *c) {
     char *line, *encoded;
 
     if(!c->opt->protocol_host) {
@@ -518,7 +570,7 @@ static void connect_client(CLI *c) {
         c->opt->protocol_host);
     fd_printf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
     if(c->opt->protocol_username && c->opt->protocol_password) {
-        if(!strcasecmp(c->opt->protocol_authentication, "NTLM")) {
+        if(!strcasecmp(c->opt->protocol_authentication, "ntlm")) {
 #if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
             ntlm(c);
 #else
@@ -541,23 +593,27 @@ static void connect_client(CLI *c) {
     }
     fd_putline(c, c->remote_fd.fd, ""); /* empty line */
     line=fd_getline(c, c->remote_fd.fd);
-    if(strlen(line)<12 || line[9]!='2') {
-        /* not "HTTP/1.0 200 Connection established" */
+    if(!is_prefix(line, "HTTP/1.0 2") && !is_prefix(line, "HTTP/1.1 2")) {
+        /* not "HTTP/1.x 2xx Connection established" */
         s_log(LOG_ERR, "CONNECT request rejected");
         do { /* read all headers */
+            str_free(line);
             line=fd_getline(c, c->remote_fd.fd);
         } while(*line);
+        str_free(line);
         longjmp(c->err, 1);
     }
     s_log(LOG_INFO, "CONNECT request accepted");
     do {
+        str_free(line);
         line=fd_getline(c, c->remote_fd.fd); /* read all headers */
     } while(*line);
+    str_free(line);
 }
 
 #if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
 
-/* 
+/*
  * NTLM code is based on the following documentation:
  * http://davenport.sourceforge.net/ntlm.html
  * http://www.innovation.ch/personal/ronald/ntlm.html
@@ -565,8 +621,8 @@ static void connect_client(CLI *c) {
 
 #define s_min(a, b) ((a)>(b)?(b):(a))
 
-static void ntlm(CLI *c) {
-    char *line, buf[BUFSIZ], *ntlm1_txt, *ntlm2_txt, *ntlm3_txt;
+NOEXPORT void ntlm(CLI *c) {
+    char *line, buf[BUFSIZ], *ntlm1_txt, *ntlm2_txt, *ntlm3_txt, *tmpstr;
     long content_length=0; /* no HTTP content */
 
     /* send Proxy-Authorization (phase 1) */
@@ -582,29 +638,39 @@ static void ntlm(CLI *c) {
     line=fd_getline(c, c->remote_fd.fd);
 
     /* receive Proxy-Authenticate (phase 2) */
-    if(line[9]!='4' || line[10]!='0' || line[11]!='7') { /* code 407 */
-        s_log(LOG_ERR, "NTLM authorization request rejected");
+    if(!is_prefix(line, "HTTP/1.0 407") && !is_prefix(line, "HTTP/1.1 407")) {
+        s_log(LOG_ERR, "Proxy-Authenticate: NTLM authorization request rejected");
         do { /* read all headers */
+            str_free(line);
             line=fd_getline(c, c->remote_fd.fd);
         } while(*line);
+        str_free(line);
         longjmp(c->err, 1);
     }
     ntlm2_txt=NULL;
     do { /* read all headers */
+        str_free(line);
         line=fd_getline(c, c->remote_fd.fd);
-        if(isprefix(line, "Proxy-Authenticate: NTLM "))
+        if(is_prefix(line, "Proxy-Authenticate: NTLM "))
             ntlm2_txt=str_dup(line+25);
-        else if(isprefix(line, "Content-Length: "))
-            content_length=atol(line+16);
+        else if(is_prefix(line, "Content-Length: ")) {
+            content_length=strtol(line+16, &tmpstr, 10);
+            if(tmpstr==line+16 || *tmpstr || content_length<0) {
+                s_log(LOG_ERR, "Proxy-Authenticate: Invalid Content-Length");
+                str_free(line);
+                longjmp(c->err, 1);
+            }
+        }
     } while(*line);
     if(!ntlm2_txt) { /* no Proxy-Authenticate: NTLM header */
         s_log(LOG_ERR, "Proxy-Authenticate: NTLM header not found");
+        str_free(line);
         longjmp(c->err, 1);
     }
 
     /* read and ignore HTTP content (if any) */
-    while(content_length) {
-        read_blocking(c, c->remote_fd.fd, buf, s_min(content_length, BUFSIZ));
+    while(content_length>0) {
+        s_read(c, c->remote_fd.fd, buf, s_min(content_length, BUFSIZ));
         content_length-=s_min(content_length, BUFSIZ);
     }
 
@@ -621,7 +687,7 @@ static void ntlm(CLI *c) {
     str_free(ntlm3_txt);
 }
 
-static char *ntlm1() {
+NOEXPORT char *ntlm1() {
     char phase1[16];
 
     memset(phase1, 0, sizeof phase1);
@@ -632,7 +698,7 @@ static char *ntlm1() {
     return base64(1, phase1, sizeof phase1); /* encode */
 }
 
-static char *ntlm3(char *username, char *password, char *phase2) {
+NOEXPORT char *ntlm3(char *username, char *password, char *phase2) {
     MD4_CTX md4;
     char *decoded; /* decoded reply from proxy */
     char phase3[146];
@@ -683,7 +749,7 @@ static char *ntlm3(char *username, char *password, char *phase2) {
     return base64(1, phase3, phase3len); /* encode */
 }
 
-static void crypt_DES(DES_cblock dst, const_DES_cblock src, DES_cblock hash) {
+NOEXPORT void crypt_DES(DES_cblock dst, const_DES_cblock src, DES_cblock hash) {
     DES_cblock key;
     DES_key_schedule sched;
 
@@ -706,7 +772,7 @@ static void crypt_DES(DES_cblock dst, const_DES_cblock src, DES_cblock hash) {
 
 #endif
 
-static char *base64(int encode, char *in, int len) {
+NOEXPORT char *base64(int encode, char *in, int len) {
     BIO *bio, *b64;
     char *out;
     int n;
