@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -42,6 +42,19 @@
 
 /**************************************** data structures */
 
+#if defined (USE_WIN32)
+#define ICON_IMAGE HICON
+#elif defined(__APPLE__)
+#define ICON_IMAGE void *
+#endif
+
+typedef enum {
+    ICON_ERROR,
+    ICON_IDLE,
+    ICON_ACTIVE,
+    ICON_NONE /* it has to be the last one */
+} ICON_TYPE;
+
 typedef enum {
     LOG_MODE_NONE,
     LOG_MODE_ERROR,
@@ -49,6 +62,12 @@ typedef enum {
     LOG_MODE_CONFIGURED
 } LOG_MODE;
 
+typedef enum {
+    FILE_MODE_READ,
+    FILE_MODE_APPEND,
+    FILE_MODE_OVERWRITE
+} FILE_MODE;
+
 typedef union sockaddr_union {
     struct sockaddr sa;
     struct sockaddr_in in;
@@ -60,19 +79,28 @@ typedef union sockaddr_union {
 #endif
 } SOCKADDR_UNION;
 
+typedef struct name_list_struct {
+    char *name;
+    struct name_list_struct *next;
+} NAME_LIST;
+
 typedef struct sockaddr_list {                          /* list of addresses */
     SOCKADDR_UNION *addr;                           /* the list of addresses */
     u16 cur;                              /* current address for round-robin */
     u16 num;                                  /* how many addresses are used */
 } SOCKADDR_LIST;
 
+#ifndef OPENSSL_NO_COMP
 typedef enum {
-    COMP_NONE, COMP_ZLIB, COMP_RLE
+    COMP_NONE, COMP_DEFLATE, COMP_ZLIB, COMP_RLE
 } COMP_TYPE;
+#endif /* OPENSSL_NO_COMP */
 
 typedef struct {
         /* some data for SSL initialization in ssl.c */
+#ifndef OPENSSL_NO_COMP
     COMP_TYPE compression;                               /* compression type */
+#endif /* OPENSSL_NO_COMP */
     char *egd_sock;                       /* entropy gathering daemon socket */
     char *rand_file;                                /* file with random data */
     int random_bytes;                       /* how many random bytes to read */
@@ -93,6 +121,12 @@ typedef struct {
     int facility;                               /* debug facility for syslog */
 #endif
     char *output_file;
+    FILE_MODE log_file_mode;
+
+        /* user interface configuraion */
+#ifdef ICON_IMAGE
+    ICON_IMAGE icon[ICON_NONE];                   /* user-specified GUI icons */
+#endif
 
         /* on/off switches */
     struct {
@@ -111,7 +145,9 @@ typedef struct {
 
 extern GLOBAL_OPTIONS global_options;
 
+#ifndef OPENSSL_NO_TLSEXT
 typedef struct servername_list_struct SERVERNAME_LIST;/* forward declaration */
+#endif
 
 typedef struct service_options_struct {
     struct service_options_struct *next;   /* next node in the services list */
@@ -140,11 +176,14 @@ typedef struct service_options_struct {
     char *cipher_list;
     char *cert;                                             /* cert filename */
     char *key;                               /* pem (priv key/cert) filename */
-    long session_timeout;
+    long session_size, session_timeout;
     long ssl_options;
     SSL_METHOD *client_method, *server_method;
     SOCKADDR_UNION sessiond_addr;
+#ifndef OPENSSL_NO_TLSEXT
+    char *sni;
     SERVERNAME_LIST *servername_list_head, *servername_list_tail;
+#endif
 #ifndef OPENSSL_NO_ECDH
     int curve;
 #endif
@@ -162,15 +201,14 @@ typedef struct service_options_struct {
     char **execargs;                     /* program arguments for local mode */
 #endif
     SOCKADDR_UNION local_addr, source_addr;
-    SOCKADDR_LIST remote_addr;
-    char *username;
-    char *remote_address;
-    char *host_name;
+    SOCKADDR_LIST connect_addr, redirect_addr;
+    NAME_LIST *connect_list, *redirect_list;
     int timeout_busy;                       /* maximum waiting for data time */
     int timeout_close;                          /* maximum close_notify time */
     int timeout_connect;                           /* maximum connect() time */
     int timeout_idle;                        /* maximum idle connection time */
     enum {FAILOVER_RR, FAILOVER_PRIO} failover;         /* failover strategy */
+    char *username;
 
         /* service-specific data for protocol.c */
     int protocol;
@@ -179,10 +217,13 @@ typedef struct service_options_struct {
     char *protocol_password;
     char *protocol_authentication;
 
-        /* service-specific data for gui.c */
+        /* service-specific data for ui_*.c */
 #ifdef USE_WIN32
-    int section_number;
+    LPTSTR file;
+    char *help;
 #endif
+    int section_number;
+    char *chain;
 
         /* on/off switches */
     struct {
@@ -208,16 +249,20 @@ typedef struct service_options_struct {
 #ifdef HAVE_OSSL_OCSP_H
         unsigned int ocsp:1;
 #endif
+        unsigned int reset:1;           /* reset sockets on error */
+        unsigned int renegotiation:1;
     } option;
 } SERVICE_OPTIONS;
 
 extern SERVICE_OPTIONS service_options;
 
+#ifndef OPENSSL_NO_TLSEXT
 struct servername_list_struct {
     char *servername;
     SERVICE_OPTIONS *opt;
     struct servername_list_struct *next;
 };
+#endif
 
 typedef enum {
     TYPE_NONE, TYPE_FLAG, TYPE_INT, TYPE_LINGER, TYPE_TIMEVAL, TYPE_STRING
@@ -250,9 +295,12 @@ typedef struct {
     struct pollfd *ufds;
     unsigned int nfds;
     unsigned int allocated;
-#else
-    fd_set irfds, iwfds, orfds, owfds;
+#else /* select */
+    fd_set *irfds, *iwfds, *ixfds, *orfds, *owfds, *oxfds;
     int max;
+#ifdef USE_WIN32
+    unsigned int allocated;
+#endif
 #endif
 } s_poll_set;
 
@@ -265,28 +313,43 @@ typedef struct disk_file {
     /* the inteface is prepared to easily implement buffering if needed */
 } DISK_FILE;
 
+    /* FD definition for client.c */
+
+typedef struct {
+    int fd; /* file descriptor */
+    int is_socket; /* file descriptor is a socket */
+} FD;
+
 /**************************************** prototypes for stunnel.c */
 
+#ifndef USE_FORK
+extern int max_clients;
 extern volatile int num_clients;
+#endif
 
-void main_initialize(char *, char *);
+void main_initialize(void);
+int main_configure(char *, char *);
+void main_cleanup(void);
+int drop_privileges(int);
 void daemon_loop(void);
 void unbind_ports(void);
 int bind_ports(void);
-#if !defined (USE_WIN32) && !defined (__vms) && !defined(USE_OS2)
-void drop_privileges(int);
+void signal_post(int);
+#if !defined(USE_WIN32) && !defined(USE_OS2)
+void child_status(void);  /* dead libwrap or 'exec' process detected */
+#endif
+void stunnel_info(int);
+
+/**************************************** prototypes for fd.c */
+
+#ifndef USE_FORK
+void get_limits(void); /* setup global max_clients and max_fds */
 #endif
 int s_socket(int, int, int, int, char *);
 int s_pipe(int [2], int, char *);
 int s_socketpair(int, int, int, int [2], int, char *);
 int s_accept(int, struct sockaddr *, socklen_t *, int, char *);
 void set_nonblock(int, unsigned long);
-void stunnel_info(int);
-void signal_post(int);
-#if !defined(USE_WIN32) && !defined(USE_OS2)
-void child_status(void);  /* dead libwrap or 'exec' process detected */
-#endif
-void die(int);
 
 /**************************************** prototypes for log.c */
 
@@ -294,16 +357,17 @@ void die(int);
 void syslog_open(void);
 void syslog_close(void);
 #endif
-void log_open(void);
+int log_open(void);
 void log_close(void);
 void log_flush(LOG_MODE);
 void s_log(int, const char *, ...)
 #ifdef __GNUC__
-    __attribute__ ((format (printf, 2, 3)));
+    __attribute__((format(printf, 2, 3)));
 #else
     ;
 #endif
-void fatal(char *, char *, int);
+void fatal_debug(char *, char *, int);
+#define fatal(a) fatal_debug((a), __FILE__, __LINE__)
 void ioerror(const char *);
 void sockerror(const char *);
 void log_error(int, int, const char *);
@@ -317,17 +381,24 @@ int pty_allocate(int *, int *, char *);
 
 extern int cli_index, opt_index;
 
-void ssl_init(void);
+int ssl_init(void);
 int ssl_configure(GLOBAL_OPTIONS *);
 
 /**************************************** prototypes for options.c */
 
-void parse_commandline(char *, char *);
-int parse_conf(char *, CONF_TYPE);
+extern char *configuration_file;
+
+int parse_commandline(char *, char *);
+int parse_conf(CONF_TYPE);
 void apply_conf(void);
 
 /**************************************** prototypes for ctx.c */
 
+typedef struct {
+    SERVICE_OPTIONS *section;
+    char pass[PEM_BUFSIZE];
+} UI_DATA;
+
 int context_init(SERVICE_OPTIONS *);
 void sslerror(char *);
 
@@ -343,6 +414,7 @@ void s_poll_init(s_poll_set *);
 void s_poll_add(s_poll_set *, int, int, int);
 int s_poll_canread(s_poll_set *, int);
 int s_poll_canwrite(s_poll_set *, int);
+int s_poll_hup(s_poll_set *, int);
 int s_poll_error(s_poll_set *, int);
 int s_poll_wait(s_poll_set *, int, int);
 
@@ -357,28 +429,31 @@ int s_poll_wait(s_poll_set *, int, int);
 #endif
 
 int set_socket_options(int, int);
-int get_socket_error(const int);
 int make_sockets(int [2]);
 
 /**************************************** prototypes for client.c */
 
-typedef struct {
-    int fd; /* file descriptor */
-    int is_socket; /* file descriptor is a socket */
-} FD;
+typedef enum {
+    RENEG_INIT, /* initial state */
+    RENEG_ESTABLISHED, /* initial handshake completed */
+    RENEG_DETECTED /* renegotiation detected */
+} RENEG_STATE;
 
 typedef struct {
+    jmp_buf err; /* exception handler needs to be 16-byte aligned on Itanium */
     SSL *ssl; /* SSL connnection */
     SERVICE_OPTIONS *opt;
-    jmp_buf err; /* exception handler */
 
     SOCKADDR_UNION peer_addr; /* peer address */
     socklen_t peer_addr_len;
+    SOCKADDR_UNION *bind_addr; /* address to bind() the socket */
+    SOCKADDR_LIST connect_addr; /* for dynamically assigned addresses */
     FD local_rfd, local_wfd; /* read and write local descriptors */
     FD remote_fd; /* remote file descriptor */
         /* IP for explicit local bind or transparent proxy */
     unsigned long pid; /* PID of the local process */
     int fd; /* temporary file descriptor */
+    RENEG_STATE reneg_state; /* used to track renegotiation attempts */
 
     /* data for transfer() function */
     char sock_buff[BUFFSIZE]; /* socket read buffer */
@@ -391,34 +466,43 @@ typedef struct {
 } CLI;
 
 CLI *alloc_client_session(SERVICE_OPTIONS *, int, int);
-void *client(void *);
+void *client_thread(void *);
+void client_main(CLI *);
 
 /**************************************** prototypes for network.c */
 
-int connect_blocking(CLI *, SOCKADDR_UNION *, socklen_t);
-void write_blocking(CLI *, int fd, void *, int);
-void read_blocking(CLI *, int fd, void *, int);
-void fdputline(CLI *, int, const char *);
-char *fdgetline(CLI *, int);
+int s_connect(CLI *, SOCKADDR_UNION *, socklen_t);
+void s_write(CLI *, int fd, void *, int);
+void s_read(CLI *, int fd, void *, int);
+void fd_putline(CLI *, int, const char *);
+char *fd_getline(CLI *, int);
 /* descriptor versions of fprintf/fscanf */
-void fdprintf(CLI *, int, const char *, ...)
+void fd_printf(CLI *, int, const char *, ...)
 #ifdef __GNUC__
-       __attribute__ ((format (printf, 3, 4)));
+    __attribute__((format(printf, 3, 4)));
 #else
-       ;
+    ;
 #endif
 
 /**************************************** prototype for protocol.c */
 
+typedef enum {
+    PROTOCOL_NONE,
+    PROTOCOL_PRE_CONNECT,
+    PROTOCOL_PRE_SSL,
+    PROTOCOL_POST_SSL
+} PROTOCOL_PHASE;
+
 int find_protocol_id(const char *);
-void protocol(CLI *c, const int);
+void protocol(CLI *, const PROTOCOL_PHASE);
 
 /**************************************** prototypes for resolver.c */
 
+void resolver_init();
 int name2addr(SOCKADDR_UNION *, char *, char *);
 int hostport2addr(SOCKADDR_UNION *, char *, char *);
-int name2addrlist(SOCKADDR_LIST *, char *, char *);
-int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+int namelist2addrlist(SOCKADDR_LIST *, NAME_LIST *, char *);
+void addrlist_dup(SOCKADDR_LIST *, const SOCKADDR_LIST *);
 char *s_ntop(SOCKADDR_UNION *, socklen_t);
 socklen_t addr_len(const SOCKADDR_UNION *);
 const char *s_gai_strerror(int);
@@ -433,9 +517,22 @@ const char *s_gai_strerror(int);
 #endif
 
 #ifdef USE_WIN32
+
 /* rename some locally shadowed declarations */
 #define getnameinfo     local_getnameinfo
-#endif /* defined USE_WIN32 */
+
+#ifndef _WIN32_WCE
+typedef int (CALLBACK * GETADDRINFO) (const char *,
+    const char *, const struct addrinfo *, struct addrinfo **);
+typedef void (CALLBACK * FREEADDRINFO) (struct addrinfo FAR *);
+typedef int (CALLBACK * GETNAMEINFO) (const struct sockaddr *, socklen_t,
+    char *, size_t, char *, size_t, int);
+extern GETADDRINFO s_getaddrinfo;
+extern FREEADDRINFO s_freeaddrinfo;
+extern GETNAMEINFO s_getnameinfo;
+#endif /* ! _WIN32_WCE */
+
+#endif /* USE_WIN32 */
 
 int getnameinfo(const struct sockaddr *, int, char *, int, char *, int, int);
 
@@ -444,22 +541,18 @@ int getnameinfo(const struct sockaddr *, int, char *, int, char *, int, int);
 /**************************************** prototypes for sthreads.c */
 
 typedef enum {
-    CRIT_CLIENTS, CRIT_SESSION, /* client.c */
-#if OPENSSL_VERSION_NUMBER<0x1000002fL
-    CRIT_SSL,                   /* client.c */
-#endif /* OpenSSL version < 1.0.0b */
-    CRIT_INET,                  /* resolver.c */
-#ifdef USE_WIN32
-    CRIT_WIN_LOG,               /* gui.c */
-#else
-    CRIT_LIBWRAP,               /* libwrap.c */
+    CRIT_CLIENTS, CRIT_SESSION, CRIT_SSL,   /* client.c */
+    CRIT_INET,                              /* resolver.c */
+#ifndef USE_WIN32
+    CRIT_LIBWRAP,                           /* libwrap.c */
 #endif
-    CRIT_SECTIONS               /* number of critical sections */
+    CRIT_LOG,                               /* log.c */
+    CRIT_SECTIONS                           /* number of critical sections */
 } SECTION_CODE;
 
 void enter_critical_section(SECTION_CODE);
 void leave_critical_section(SECTION_CODE);
-void sthreads_init(void);
+int sthreads_init(void);
 unsigned long stunnel_process_id(void);
 unsigned long stunnel_thread_id(void);
 int create_client(int, int, CLI *, void *(*)(void *));
@@ -485,41 +578,12 @@ void _endthread(void);
 void stack_info(int);
 #endif
 
-/**************************************** prototypes for gui.c */
-
-typedef struct {
-    SERVICE_OPTIONS *section;
-    char pass[PEM_BUFSIZE];
-} UI_DATA;
-
-#ifdef USE_WIN32
-void win_log(char *);
-void win_exit(const int);
-void win_newconfig();
-void win_newcert(SSL *, SERVICE_OPTIONS *);
-int passwd_cb(char *, int, int, void *);
-#ifdef HAVE_OSSL_ENGINE_H
-int pin_cb(UI *, UI_STRING *);
-#endif
-
-#ifndef _WIN32_WCE
-typedef int (CALLBACK * GETADDRINFO) (const char *,
-    const char *, const struct addrinfo *, struct addrinfo **);
-typedef void (CALLBACK * FREEADDRINFO) (struct addrinfo FAR *);
-typedef int (CALLBACK * GETNAMEINFO) (const struct sockaddr *, socklen_t,
-    char *, size_t, char *, size_t, int);
-extern GETADDRINFO s_getaddrinfo;
-extern FREEADDRINFO s_freeaddrinfo;
-extern GETNAMEINFO s_getnameinfo;
-#endif /* ! _WIN32_WCE */
-#endif /* USE_WIN32 */
-
 /**************************************** prototypes for file.c */
 
 #ifndef USE_WIN32
 DISK_FILE *file_fdopen(int);
 #endif
-DISK_FILE *file_open(char *, int);
+DISK_FILE *file_open(char *, FILE_MODE mode);
 void file_close(DISK_FILE *);
 int file_getline(DISK_FILE *, char *, int);
 int file_putline(DISK_FILE *, char *);
@@ -531,13 +595,13 @@ LPSTR tstr2str(const LPTSTR);
 
 /**************************************** prototypes for libwrap.c */
 
-void libwrap_init();
+int libwrap_init();
 void libwrap_auth(CLI *, char *);
 
 /**************************************** prototypes for str.c */
 
 void str_init();
-void str_canary();
+void str_canary_init();
 void str_cleanup();
 void str_stats();
 void *str_alloc_debug(size_t, char *, int);
@@ -547,16 +611,38 @@ void *str_realloc_debug(void *, size_t, char *, int);
 void str_detach_debug(void *, char *, int);
 #define str_detach(a) str_detach_debug((a), __FILE__, __LINE__)
 void str_free_debug(void *, char *, int);
-#define str_free(a) str_free_debug((a), __FILE__, __LINE__)
-char *str_dup(const char *);
+#define str_free(a) str_free_debug((a), __FILE__, __LINE__), (a)=NULL
+char *str_dup_debug(const char *, char *, int);
+#define str_dup(a) str_dup_debug((a), __FILE__, __LINE__)
 char *str_vprintf(const char *, va_list);
 char *str_printf(const char *, ...)
 #ifdef __GNUC__
-    __attribute__ ((format (printf, 1, 2)));
+    __attribute__((format(printf, 1, 2)));
 #else
     ;
 #endif
 
+/**************************************** prototypes for ui_*.c */
+
+void ui_new_config(void);
+void ui_new_chain(const int);
+void ui_clients(const int);
+
+void ui_new_log(const char *);
+#ifdef USE_WIN32
+void message_box(const LPSTR, const UINT);
+#endif /* USE_WIN32 */
+
+int passwd_cb(char *, int, int, void *);
+#ifdef HAVE_OSSL_ENGINE_H
+int pin_cb(UI *, UI_STRING *);
+#endif
+
+#ifdef ICON_IMAGE
+ICON_IMAGE load_icon_default(ICON_TYPE);
+ICON_IMAGE load_icon_file(const char *);
+#endif
+
 #endif /* defined PROTOTYPES_H */
 
 /* end of prototypes.h */
