@@ -1,24 +1,24 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
  *   Free Software Foundation; either version 2 of the License, or (at your
  *   option) any later version.
- * 
+ *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  *   See the GNU General Public License for more details.
- * 
+ *
  *   You should have received a copy of the GNU General Public License along
  *   with this program; if not, see <http://www.gnu.org/licenses>.
- * 
+ *
  *   Linking stunnel statically or dynamically with other modules is making
  *   a combined work based on stunnel. Thus, the terms and conditions of
  *   the GNU General Public License cover the whole combination.
- * 
+ *
  *   In addition, as a special exception, the copyright holder of stunnel
  *   gives you permission to combine stunnel with free software programs or
  *   libraries that are released under the GNU LGPL and with code included
@@ -26,7 +26,7 @@
  *   modified versions of such code, with unchanged license). You may copy
  *   and distribute such a system following the terms of the GNU GPL for
  *   stunnel and the licenses of the other code concerned.
- * 
+ *
  *   Note that people who make modified versions of stunnel are not obligated
  *   to grant this special exception for their modified versions; it is their
  *   choice whether to do so. The GNU General Public License gives permission
@@ -40,6 +40,9 @@
 
 /**************************************** prototypes */
 
+NOEXPORT int name2addrlist(SOCKADDR_LIST *, char *, char *);
+NOEXPORT int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+
 #ifndef HAVE_GETADDRINFO
 
 #ifndef EAI_MEMORY
@@ -52,11 +55,11 @@
 #define EAI_SERVICE 8
 #endif
 
-#ifdef USE_WIN32
-/* rename some locally shadowed declarations */
+/* rename some potentially locally shadowed declarations */
 #define getaddrinfo     local_getaddrinfo
 #define freeaddrinfo    local_freeaddrinfo
-#else /* defined USE_WIN32 */
+
+#ifndef HAVE_STRUCT_ADDRINFO
 struct addrinfo {
     int ai_flags;
     int ai_family;
@@ -67,16 +70,52 @@ struct addrinfo {
     char *ai_canonname;
     struct addrinfo *ai_next;
 };
-#endif /* defined USE_WIN32 */
+#endif
 
-static int getaddrinfo(const char *, const char *,
+NOEXPORT int getaddrinfo(const char *, const char *,
     const struct addrinfo *, struct addrinfo **);
-static int alloc_addresses(struct hostent *, const struct addrinfo *,
+NOEXPORT int alloc_addresses(struct hostent *, const struct addrinfo *,
     u_short port, struct addrinfo **, struct addrinfo **);
-static void freeaddrinfo(struct addrinfo *);
+NOEXPORT void freeaddrinfo(struct addrinfo *);
 
 #endif /* !defined HAVE_GETADDRINFO */
 
+/**************************************** resolver initialization */
+
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
+GETADDRINFO s_getaddrinfo;
+FREEADDRINFO s_freeaddrinfo;
+GETNAMEINFO s_getnameinfo;
+#endif
+
+void resolver_init() {
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
+    HINSTANCE handle;
+
+    handle=LoadLibrary("ws2_32.dll"); /* IPv6 in Windows XP or higher */
+    if(handle) {
+        s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
+        s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
+        s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
+        if(s_getaddrinfo && s_freeaddrinfo && s_getnameinfo)
+            return; /* IPv6 detected -> OK */
+        FreeLibrary(handle);
+    }
+    handle=LoadLibrary("wship6.dll"); /* experimental IPv6 for Windows 2000 */
+    if(handle) {
+        s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
+        s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
+        s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
+        if(s_getaddrinfo && s_freeaddrinfo && s_getnameinfo)
+            return; /* IPv6 detected -> OK */
+        FreeLibrary(handle);
+    }
+    s_getaddrinfo=NULL;
+    s_freeaddrinfo=NULL;
+    s_getnameinfo=NULL;
+#endif
+}
+
 /**************************************** stunnel resolver API */
 
 int name2addr(SOCKADDR_UNION *addr, char *name, char *default_host) {
@@ -107,7 +146,15 @@ int hostport2addr(SOCKADDR_UNION *addr, char *hostname, char *portname) {
     return retval;
 }
 
-int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_host) {
+int namelist2addrlist(SOCKADDR_LIST *addr_list, NAME_LIST *name_list, char *default_host) {
+    /* recursive implementation to reverse the list */
+    if(!name_list)
+        return 0;
+    return namelist2addrlist(addr_list, name_list->next, default_host) +
+        name2addrlist(addr_list, name_list->name, default_host);
+}
+
+NOEXPORT int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_host) {
     char *tmp, *hostname, *portname;
     int retval;
 
@@ -146,10 +193,10 @@ int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_host) {
     return retval;
 }
 
-int hostport2addrlist(SOCKADDR_LIST *addr_list,
+NOEXPORT int hostport2addrlist(SOCKADDR_LIST *addr_list,
         char *hostname, char *portname) {
     struct addrinfo hints, *res=NULL, *cur;
-    int err;
+    int err, retries=0;
 
     memset(&hints, 0, sizeof hints);
 #if defined(USE_IPv6) || defined(USE_WIN32)
@@ -159,15 +206,15 @@ int hostport2addrlist(SOCKADDR_LIST *addr_list,
 #endif
     hints.ai_socktype=SOCK_STREAM;
     hints.ai_protocol=IPPROTO_TCP;
-    do {
+    for(;;) {
         err=getaddrinfo(hostname, portname, &hints, &res);
         if(err && res)
             freeaddrinfo(res);
-        if(err==EAI_AGAIN) {
-            s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
-            sleep(1);
-        }
-    } while(err==EAI_AGAIN);
+        if(err!=EAI_AGAIN || ++retries>=3)
+            break;
+        s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
+        sleep(1);
+    }
     switch(err) {
     case 0:
         break; /* success */
@@ -182,7 +229,7 @@ int hostport2addrlist(SOCKADDR_LIST *addr_list,
 
     /* copy the list of addresses */
     for(cur=res; cur; cur=cur->ai_next) {
-        if(cur->ai_addrlen>sizeof(SOCKADDR_UNION)) {
+        if(cur->ai_addrlen>(int)sizeof(SOCKADDR_UNION)) {
             s_log(LOG_ERR, "INTERNAL ERROR: ai_addrlen value too big");
             freeaddrinfo(res);
             return 0; /* no results */
@@ -196,6 +243,14 @@ int hostport2addrlist(SOCKADDR_LIST *addr_list,
     return addr_list->num; /* ok - return the number of addresses */
 }
 
+void addrlist_dup(SOCKADDR_LIST *dst, const SOCKADDR_LIST *src) {
+    memcpy(dst, src, sizeof(SOCKADDR_LIST));
+    if(src->addr) {
+        dst->addr=str_alloc(src->num*sizeof(SOCKADDR_UNION));
+        memcpy(dst->addr, src->addr, src->num*sizeof(SOCKADDR_UNION));
+    }
+}
+
 char *s_ntop(SOCKADDR_UNION *addr, socklen_t addrlen) {
     int err;
     char *host, *port, *retval;
@@ -236,7 +291,7 @@ socklen_t addr_len(const SOCKADDR_UNION *addr) {
 /* implementation is limited to functionality needed by stunnel */
 
 #ifndef HAVE_GETADDRINFO
-static int getaddrinfo(const char *node, const char *service,
+NOEXPORT int getaddrinfo(const char *node, const char *service,
         const struct addrinfo *hints, struct addrinfo **res) {
     struct hostent *h;
 #ifndef _WIN32_WCE
@@ -266,8 +321,6 @@ static int getaddrinfo(const char *node, const char *service,
 
     /* allocate addrlist structure */
     ai=str_alloc(sizeof(struct addrinfo));
-    if(!ai)
-        return EAI_MEMORY;
     if(hints)
         memcpy(ai, hints, sizeof(struct addrinfo));
 
@@ -276,10 +329,6 @@ static int getaddrinfo(const char *node, const char *service,
     ai->ai_family=AF_INET6;
     ai->ai_addrlen=sizeof(struct sockaddr_in6);
     ai->ai_addr=str_alloc(ai->ai_addrlen);
-    if(!ai->ai_addr) {
-        str_free(ai);
-        return EAI_MEMORY;
-    }
     ai->ai_addr->sa_family=AF_INET6;
     if(inet_pton(AF_INET6, node,
             &((struct sockaddr_in6 *)ai->ai_addr)->sin6_addr)>0) {
@@ -287,10 +336,6 @@ static int getaddrinfo(const char *node, const char *service,
     ai->ai_family=AF_INET;
     ai->ai_addrlen=sizeof(struct sockaddr_in);
     ai->ai_addr=str_alloc(ai->ai_addrlen);
-    if(!ai->ai_addr) {
-        str_free(ai);
-        return EAI_MEMORY;
-    }
     ai->ai_addr->sa_family=AF_INET;
     ((struct sockaddr_in *)ai->ai_addr)->sin_addr.s_addr=inet_addr(node);
     if(((struct sockaddr_in *)ai->ai_addr)->sin_addr.s_addr+1) {
@@ -307,7 +352,7 @@ static int getaddrinfo(const char *node, const char *service,
     *res=NULL;
     ai=NULL;
     enter_critical_section(CRIT_INET);
-#if defined(USE_IPv6) && !defined(USE_WIN32)
+#ifdef HAVE_GETHOSTBYNAME2
     h=gethostbyname2(node, AF_INET6);
     if(h) /* some IPv6 addresses found */
         alloc_addresses(h, hints, port, res, &ai); /* ignore the error */
@@ -332,7 +377,7 @@ static int getaddrinfo(const char *node, const char *service,
     return retval;
 }
 
-static int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
+NOEXPORT int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
         u_short port, struct addrinfo **head, struct addrinfo **tail) {
     int i;
     struct addrinfo *ai;
@@ -340,8 +385,6 @@ static int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
     /* copy addresses */
     for(i=0; h->h_addr_list[i]; i++) {
         ai=str_alloc(sizeof(struct addrinfo));
-        if(!ai)
-            return EAI_MEMORY;
         if(hints)
             memcpy(ai, hints, sizeof(struct addrinfo));
         ai->ai_next=NULL; /* just in case */
@@ -357,8 +400,6 @@ static int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
         if(h->h_addrtype==AF_INET6) {
             ai->ai_addrlen=sizeof(struct sockaddr_in6);
             ai->ai_addr=str_alloc(ai->ai_addrlen);
-            if(!ai->ai_addr)
-                return EAI_MEMORY;
             memcpy(&((struct sockaddr_in6 *)ai->ai_addr)->sin6_addr,
                 h->h_addr_list[i], h->h_length);
         } else
@@ -366,8 +407,6 @@ static int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
         {
             ai->ai_addrlen=sizeof(struct sockaddr_in);
             ai->ai_addr=str_alloc(ai->ai_addrlen);
-            if(!ai->ai_addr)
-                return EAI_MEMORY;
             memcpy(&((struct sockaddr_in *)ai->ai_addr)->sin_addr,
                 h->h_addr_list[i], h->h_length);
         }
@@ -378,7 +417,7 @@ static int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
     return 0; /* success */
 }
 
-static void freeaddrinfo(struct addrinfo *current) {
+NOEXPORT void freeaddrinfo(struct addrinfo *current) {
     struct addrinfo *next;
 
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
